from .ctrls import (CS, EraseMode, move, home, clear, erase, hide_cursor,
                    save_pos, alternate_screen, nestable_save_pos_possible)
from .sgr import SGR, len_diff, text, fg, bg, sgr_string, sgr_print
from .spec import set_attributes_delimiters, parse

__version__ = '0.3.0'

__all__ = ['CS',
           'SGR',
           'EraseMode',
           'erase',
           'move',
           'home',
           'clear',
           'hide_cursor',
           'nestable_save_pos_possible',
           'save_pos',
           'alternate_screen',
           'fg',
           'bg',
           'sgr_string',
           'sgr_print',
           'text',
           'len_diff',
           'set_attributes_delimiters',
           'parse',
           ]



from contextlib import contextmanager
from enum import Enum
from functools import partial
from types import new_class

import term


_CTRLSEQS = dict(
    ICH='\033[%s@',     # insert character(s)
    CUU='\033[%sA',     # cusor up
    CUD='\033[%sB',     # cursor down
    CUF='\033[%sC',     # cursor foreward
    CUB='\033[%sD',     # cursor backward
    CNL='\033[%sE',     # cursor next line
    CPL='\033[%sF',     # cursor preceding line
    CHA='\033[%sG',     # cursor character absolute
    CUP='\033[%sH',     # cursor position
    CHT='\033[%sI',     # cursor forward tabulation
    ED='\033[%sJ',      # erase in display
    EL='\033[%sK',      # erase in line
    IL='\033[%sL',      # insert line(s)
    DL='\033[%sM',      # delete line(s)
    DCH='\033[%sP',     # delete character(s)
    SU='\033[%sS',      # scroll up
    SD='\033[%sT',      # scroll down
    ECH='\033[%sX',     # erase character(s)
    CBT='\033[%sZ',     # cursor backward tabulation
    HPA='\033[%s`',     # horizontal postion absolute
    VPA='\033[%sd',     # vertical position absolute
    SGR='\033[%sm',     # select graphic rendition
    SCP='\033[s',       # save cursor position
    RCP='\033[u',       # restore cursor position
    SCU='\033[?25h',    # show cursor
    HCU='\033[?25l',    # hide cursor
    EAS='\033[?1049h',  # enable alternate screen buffer
    DAS='\033[?1049l',  # disable alternate screen buffer
    RIS='\033c',        # reset screen to initial state
)


class _CtrlSeq:
    def __init__(self, name, value):
        self._name = name
        self._value = value

    def __mod__(self, other):
        if '%s' in self._value:
            if isinstance(other, int):
                return self._value % int(other)
            if isinstance(other, tuple):
                lst = []
                for x in other:
                    if isinstance(x, tuple):
                        lst.extend(x)
                    else:
                        lst.append(x)
                if all(isinstance(x, int) for x in lst):
                    return self._value % ';'.join(map(str, map(int, lst)))
            raise TypeError('Arguments must be of type int or a tuple of ints')
        else:
            return self._value

    def __repr__(self):
        return '<%s: %r>' % (self._name, self._value)

    def __str__(self):
        if '%s' in self._value:
            return self._value % ''
        else:
            return self._value

    def __add__(self, other):
        return str(self) + str(other)

    def __radd__(self, other):
        return str(other) + str(self)


def _ctrlseqs_cb(ns):
    for k, v in _CTRLSEQS.items():
        ns[k] = _CtrlSeq('CS.' + k, v)


CS = new_class('CS', exec_body=_ctrlseqs_cb)
CS.__doc__ = 'Command sequences.'

_print = partial(print, end='', sep='', flush=True)


def move(row=0, col=0, rel=True):
    """Move the cursor.

    The cursor can be moved relative (``rel=True``) or absolute (``rel=False``).
    If ``row=0`` the cursor will be moved within the current row; if ``col=0``
    the cursor will be moved in the current column. Absolute ``row`` and ``col``
    values start at 1 and cannot be negative. For relative movements negative
    numbers mean to the left or up and positive numbers move the cursor to the
    right or down. If only one of ``row`` or ``col`` is relative and the other
    is absolute, the ``rel`` parameter has to be a tuple: ``rel=(True, False)``
    means relative row and absolute column movement and ``rel=(False, True)``
    absolute row and relative column movement.

    ::

       move(col=2)        # move cursor two columns to the right
                          # and leave the row unchanged
       move(1, 1, False)  # move cursor to the top left corner
       move(-5, 1, (True, False)  # move cursor to the beginning of the
                                  # line five rows up

    :param int row: the row
    :param int col: the column
    :param rel: if ``True`` the cursor will be moved relative to the
                current position
    :type rel: bool or tuple(bool, bool)
    :raises TypeError: if row or col are not of type ``int``
    :raises ValueError: if not at least one of ``row`` or ``col`` is not 0,
                        or if ``rel=False`` and ``row`` or ``col`` are
                        negative
    """
    if not (isinstance(row, int) and isinstance(col, int)):
        raise TypeError('Arguments "row" and "col" must be of type int')
    if row == 0 and col == 0:
        raise ValueError('At least one argument "row/col" != 0 required')
    if isinstance(rel, tuple):
        rel_row, rel_col = map(bool, rel)
    else:
        rel_row = rel_col = bool(rel)
    if (not rel_row and row < 0) or (not rel_col and col < 0):
        raise ValueError('Absolute "row/col" must be positive integers')
    if row and col and rel_row == rel_col is False:
        _print(CS.CUP % (row, col))
    else:
        if row:
            if rel_row:
                _print(CS.CUD % row if row > 0 else CS.CUU % abs(row))
            else:
                _print(CS.VPA % row)
        if col:
            if rel_col:
                _print(CS.CUF % col if col > 0 else CS.CUB % abs(col))
            else:
                _print(CS.HPA % col)


def home():
    """Move the cursor to the top left corner of the screen."""
    _print(CS.CUP)


def clear(reset=False):
    """Clear screen.

    Erase entire screen and move the cursor to the top left corner
    of the screen.

    :param bool reset: if set to `True` the line buffer will be erased too
    """
    erase(EraseMode.SCRN_ALL)
    home()
    if reset:
        erase(EraseMode.SCRN_LINES)


class EraseMode(Enum):
    """Mode parameter for :func:`erase`."""

    SCRN_END = 0    #: From cursor to end of screen
    SCRN_BEGIN = 1  #: From cursor to beginning of screen
    SCRN_ALL = 2    #: Entire screen
    SCRN_LINES = 3  #: All lines in scrollback buffer
    LINE_END = 4    #: From cursor to end of line
    LINE_BEGIN = 5  #: From cursor to beginning of line
    LINE_ALL = 6    #: Entire line


def erase(mode):
    """Erase all or part of line/screen (cursor position does not change).

    :param mode: what to erase
    :type mode: EraseMode
    """
    if not isinstance(mode, EraseMode):
        raise TypeError('Argument "mode" must be an EraseMode enum member')
    if mode.value <= 3:
        _print(CS.ED % mode.value)
    else:
        _print(CS.EL % (mode.value - 4))


@contextmanager
def hide_cursor():
    """Hide the cursor.

    This function is a context manager for use in ``with`` statements.
    It hides the cursor when the context is entered and shows it again
    when the context is exited.
    """
    _print(CS.HCU)
    try:
        yield
    finally:
        _print(CS.SCU)


def nestable_save_pos_possible():
    """Return if nestable :func:`save_pos` context is supported.

    :return: True if supported
    :rtype: bool
    """
    return term.getyx() != (0, 0)


@contextmanager
def save_pos(nestable=False):
    """Save the cursor position.

    This function is a context manager for use in ``with`` statements.
    It saves the cursor position when the context is entered and restores the
    cursor to the saved position when the context is exited.

    If ``nestable=False`` the ANSI control sequences ``ESC[s: SCP - Save Cursor
    Position`` and ``ESC[u: RCP - Restore Cursor Position`` will be used. By
    doing so only the saved cursor position of the innermost context will be
    restored.

    If ``nestable=True`` and :func:`nestable_save_pos_possible` returns
    ``True``, each time the context is entered the position is determined by
    calling ``getyx()`` from the
    `term package <https://pypi.org/project/term/>`_.
    If the terminal does not support the control sequence
    ``ESC[6n: DSR – Device Status Report`` this is not possible.

    :param bool nestable: wether a nestable context should be used
    :raises RuntimeError: if ``nestable=True`` and nestable context not possible
    """
    if nestable:
        pos = term.getyx()
        if pos == (0, 0):
            raise RuntimeError('Nestable save_pos context not supported')
    else:
        _print(CS.SCP)
    try:
        yield
    finally:
        if nestable:
            move(*pos, False)
        else:
            _print(CS.RCP)


@contextmanager
def alternate_screen():
    """Use the `alternate screen buffer`_.

    .. _alternate screen buffer: https://invisible-island.net/xterm/ctlseqs/
       ctlseqs.html#h2-The-Alternate-Screen-Buffer

    This function is a context manager for use in ``with`` statements.
    It switches to the alternate screen buffer when the context is entered and
    back to the normal screen buffer when the context is exited.
    """
    _print(CS.EAS)
    try:
        yield
    finally:
        _print(CS.DAS)



import re
from contextlib import suppress

from .ctrls import CS
from .sgr import SGR, fg, bg

_ATTRS_GRP_PATTERN = '(?P<attrs>.*?)'
_DFLT_ATTRS_START = '#['
_DFLT_ATTRS_END = ']'
_ATTRS = {'bd': SGR.Bold,
          'ft': SGR.Faint,
          'it': SGR.Italic,
          'ul': SGR.Underlined,
          'bk': SGR.Blink,
          'iv': SGR.Inverse,
          'co': SGR.Crossedout,
          'ol': SGR.Overlined,
          '': SGR.Reset,
          'reset': SGR.Reset}


def set_attributes_delimiters(start=_DFLT_ATTRS_START, end=_DFLT_ATTRS_END):
    """Set the delimiters for attributes.

    See: :ref:`specification string <ref-intro-spec>`.

    :param str start: start delimiter
    :param str end: end delimiter
    :raise TypeError: if one of the delimiters is not a string
    :raise ValueError: if one of the delimiters is an empty string
    """
    global _attrs_re
    if not (start and end):
        raise ValueError(
            'at least one delimiter %r/%r not allowed' % (start, end))
    _attrs_re = re.compile(re.escape(start) + _ATTRS_GRP_PATTERN +
                           re.escape(end))


set_attributes_delimiters()  # set the defaults


def parse(spec, strict=False):
    """Parse the :ref:`specification string <ref-intro-spec>`.

    :param str spec: the specification string
    :param bool strict: if ``False`` unknown attributes will be ignored
    :return: string with SGR control sequences
    :rtype: str
    :raises ValueError: if ``strict=True`` and an unkown attribute
                        is encountered
    """
    tokens = []
    pos = 0
    for m in _attrs_re.finditer(spec):
        start, end = m.span()
        if pos != start:
            tokens.append(spec[pos:start])
        pos = end
        params = _get_params(m['attrs'], strict)
        if params:
            tokens.append(CS.SGR % params)
    if pos != len(spec):
        tokens.append(spec[pos:])
    return ''.join(tokens)


def _get_params(s, strict):
    params = []
    for x in map(str.strip, s.lower().split(';')):
        if x.startswith('fg') or x.startswith('bg'):
            params.append(_get_color(x))
            continue
        if x.startswith('not'):
            not_ = True
            x = x[3:].lstrip()
        else:
            not_ = False
        attr = _ATTRS.get(x)
        if attr is not None:
            params.append(~attr if not_ else attr)
        else:
            params.append(None)
    if strict and None in params:
        raise ValueError('Error in attributes: "%s"' % s)
    return tuple(param for param in params if param is not None)


def _get_color(s):
    is_fg = s.startswith('fg')
    s = s[2:].lstrip()
    if not s or s == 'default':
        return SGR.FgDefault if is_fg else SGR.BgDefault
    with suppress(Exception):
        if s.startswith('#'):
            if len(s) in (4, 7):
                if len(s) == 4:
                    s = ''.join(x + x for x in s[1:])
                else:
                    s = s[1:]
                t = tuple(int(s[i:i + 2], 16) for i in range(0, len(s), 2))
                if all(0 <= n for n in t):
                    return fg(*t) if is_fg else bg(*t)
        elif ',' in s:
            t = tuple(int(x) for x in s.split(',', 2) if 0 <= int(x) <= 255)
            if len(t) == 3:
                return fg(*t) if is_fg else bg(*t)
        elif s.isdigit():
            if 0 <= int(s) <= 255:
                return fg(int(s)) if is_fg else bg(int(s))
        else:
            if s.startswith('bright'):
                bright = True
                s = s[6:].lstrip()
            else:
                bright = False
            clr = getattr(SGR, ('Fg' if is_fg else 'Bg') + s.capitalize())
            return ~clr if bright else clr
    return None



import re
from contextlib import suppress
from functools import lru_cache
from types import new_class

from .ctrls import CS

_SGRRE = re.compile(r'\x1b\[[\d;]*m')

_SGR_ATTRS = dict(
    Bold=(1, 22),
    Faint=(2, 22),
    Italic=(3, 23),
    Underlined=(4, 24),
    Blink=(5, 25),
    Inverse=(7, 27),
    Crossedout=(9, 29),
    Overlined=(53, 55),
    Reset=(0, 0),
    FgBlack=(30, 90),
    FgRed=(31, 91),
    FgGreen=(32, 92),
    FgYellow=(33, 93),
    FgBlue=(34, 94),
    FgMagenta=(35, 95),
    FgCyan=(36, 96),
    FgWhite=(37, 97),
    FgDefault=(39, 39),
    BgBlack=(40, 100),
    BgRed=(41, 101),
    BgGreen=(42, 102),
    BgYellow=(43, 103),
    BgBlue=(44, 104),
    BgMagenta=(45, 105),
    BgCyan=(46, 106),
    BgWhite=(47, 107),
    BgDefault=(49, 49),
)


class _SGRParam:
    pass


class _SGRAttr(int, _SGRParam):
    def __new__(cls, name, value, invert):
        obj = int.__new__(cls, value)
        obj._name = name
        obj._invert = int.__new__(cls, invert)
        obj._invert._name = '~' + name
        obj._invert._invert = obj
        return obj

    def __invert__(self):
        return self._invert

    def __repr__(self):
        return '<%s: %d>' % (self._name, self)

    def __str__(self):
        return '%s' % self._name


def _sgr_cb(ns):
    for k, v in _SGR_ATTRS.items():
        ns[k] = _SGRAttr('SGR.' + k, *v)


SGR = new_class('SGR', exec_body=_sgr_cb)
SGR.__doc__ = 'SGR attributes.'


class _SGRColor(tuple, _SGRParam):
    def __new__(cls, name, *args):
        obj = tuple.__new__(cls, args)
        obj._name = name
        return obj

    def __repr__(self):
        return '<SGR.%sColor: %s>' % (self._name, tuple(self))

    def __str__(self):
        return 'SGR.%sColor%s' % (self._name, self[2:])


def _color(*args):
    with suppress(Exception):
        if len(args) == 1:
            if isinstance(args[0], int) and 0 <= args[0] <= 255:
                return 5, args[0]
            elif len(args[0]) in (4, 7) and args[0].startswith('#'):
                if len(args[0]) == 4:
                    s = ''.join(x + x for x in args[0][1:])
                else:
                    s = args[0][1:]
                return (2,) + tuple(int(s[i:i + 2], 16)
                                    for i in range(0, len(s), 2))
        elif len(args) == 3 and all(0 <= x <= 255 for x in args):
            return (2,) + args
    raise ValueError('Invalid arguments for color: %s' % str(args))


def fg(*args):
    """Return SGR parameters for foreground :ref:`color <ref-intro-colors>`.

    :param args: arguments for color
    :return: color for use with :attr:`CS.SGR` or :func:`sgr_print`
    """
    return _SGRColor('Fg', 38, *_color(*args))


def bg(*args):
    """Return SGR parameters for background :ref:`color <ref-intro-colors>`.

    :param args: arguments for color
    :return: color for use with :attr:`CS.SGR` or :func:`sgr_print`
    """
    return _SGRColor('Bg', 48, *_color(*args))


@lru_cache()
def text(s):
    """Return the plain text without the SGR control sequences.

    :param str s: the string with SGR control sequences
    :return: plain text
    :rtype: str
    """
    return ''.join(_SGRRE.split(s))


def len_diff(s):
    r"""Return the difference between ``len(s)`` and ``len(text(s))``.

    Useful for adjusting the text on the screen::

       s = sgr_string(SGR.Bold, SGR.FgRed, 'ABC')
       w = 20
       print(repr(s))
       print(text(s))
       print('1234567890' * 2)  # numbering 20 columns
       print(s.center(w, '.'))
       print(s.center(w + len_diff(s), '.'))
       print('%*s' % (w, s))
       print('%*s' % (w + len_diff(s), s))
       print('{:.^{width}}'.format(s, width=w))
       print('{:.^{width}}'.format(s, width=w + len_diff(s)))

    .. raw:: html

       <pre style="color:#FFFFFF;background-color:#000000">
       &apos;\x1b[1;31mABC\x1b[0m&apos;
       ABC
       12345678901234567890
       ...<font color="#FF5555"><b>ABC</b></font>...
       .........<font color="#FF5555"><b>ABC</b></font>........
             <font color="#FF5555"><b>ABC</b></font>
                        <font color="#FF5555"><b>ABC</b></font>
       ...<font color="#FF5555"><b>ABC</b></font>...
       ........<font color="#FF5555"><b>ABC</b></font>.........
       </pre>

    :param str s: the string with SGR control sequences
    :return: difference of the lengths of ``s`` and the visible text
    :rtype: int
    """
    return len(s) - len(text(s))


def sgr_string(*args, reset=True):
    r"""Build a string with SGR control sequences.

    Those arguments that are :class:`SGR` attributes or the result of
    the :func:`fg` or :func:`bg` functions will be passed as parameters to
    :attr:`CS.SGR`, all others will be converted to strings. Than
    all elements will be concatenated and returned as a string.

    >>> CS.SGR % (SGR.Bold, SGR.FgRed) + 'ABC' + CS.SGR % SGR.Reset
    '\x1b[1;31mABC\x1b[0m'
    >>> sgr_string(SGR.Bold, SGR.FgRed, 'ABC')
    '\x1b[1;31mABC\x1b[0m'

    :param args: arguments
    :param bool reset: if ``True`` all SGR attributes will be reset to normal
                       at the end of the string
    :return: a string
    :rtype: str
    """
    r = []
    p = []
    for arg in args:
        if isinstance(arg, _SGRParam):
            p.append(arg)
        else:
            if p:
                r.append(CS.SGR % tuple(p))
                p.clear()
            r.append(arg)
    if reset:
        p.append(SGR.Reset)
    if p:
        r.append(CS.SGR % tuple(p))
    return ''.join(map(str, r))


def sgr_print(*args, reset=True, end='\n', flush=False):
    """Print a string with SGR control sequences.

    See function :func:`sgr_string`.

    :param args: arguments (see: :func:`sgr_string`)
    :param bool reset: if ``True`` all SGR attributes will be reset to normal
                       at the end of the string
    :param str end: see built-in function :func:`print`
    :param bool flush: see built-in function :func:`print`
    """
    print(sgr_string(*args, reset=reset), sep='', end=end, flush=flush)
